/*
  dxlib.cpp
  Dynamixel Protocol V.1 Library for DXSHIELD

  Copyright (c) BestTechnology CO.,LTD. 2017
*/

#include "dxlib.h"


DXLIB::DXLIB (bool ss, HardwareSerial *hws, uint8_t rxpin, uint8_t txpin) {
  UseSoftSerial = ss;
  txp = txpin;
  rxp = rxpin;
}

void DXLIB::begin (long baud) {
  if (UseSoftSerial) {
    _dxif = new SoftwareSerial (rxp, txp);
    _dxif->begin (baud);
    _dxif->setTimeout (20);
  } else {
    _HardSerial->begin (baud);
    _HardSerial->setTimeout (20);
  }
}

DXLIB::~DXLIB() {
  if (UseSoftSerial) {
    _dxif->end();
    _dxif = NULL;
  } else {
    _HardSerial->end();
  }
}

bool DXLIB::active (void) {
  if (UseSoftSerial) {
    return (_dxif != NULL);
  } else {
    if (Serial) return true;
    else return false;
  }
}

void DXLIB::PurgeRXBuff (void) {
  if (UseSoftSerial) {
    while (_dxif->available()) _dxif->read();
  } else {
    while (_HardSerial->available()) _HardSerial->read();
  }
}

void DXLIB::puts (uint8_t *buf, int len) {
  if (UseSoftSerial) _dxif->write (buf, len);
  else _HardSerial->write (buf, len);
}

void DXLIB::putch (uint8_t c) {
  if (UseSoftSerial) _dxif->write (c);
  else _HardSerial->write (c);
}

int DXLIB::gets (uint8_t *buf, int len) {
  return (UseSoftSerial) ? _dxif->readBytes (buf, len) : _HardSerial->readBytes (buf, len);
}

void DXLIB::flush (void) {
  if (UseSoftSerial) _dxif->flush();
  else _HardSerial->flush();
}

bool DXLIB::TxPacket (uint8_t id, uint8_t inst, uint8_t *param, uint16_t len) {
  if (!active()) return false;
  if (id > 254) return false;
  uint8_t sum = 0, buf[5] = {0xff, 0xff, id, len + 2, inst};
  PurgeRXBuff();
  for (int i = 2; i < 5; i++) sum += buf[i];
  puts (buf, 5);
  if (len > 0) {
    for (int i = 0; i < len; i++) sum += param[i];
    puts (param, len);
  }
  sum = ~sum;
  puts (&sum, 1);
  flush();
  return true;
}

bool DXLIB::RxPacket (uint8_t *rdata, uint16_t rdatasize, uint8_t *rid, uint16_t *rlen, uint8_t *err) {
  if (!active()) return false;
  if (rdata == NULL) return false;
  if (rlen == NULL) return false;
  uint8_t buf[5];
  if (gets (buf, 5) == 5) {
    if ((* (uint16_t *)&buf[0] == 0xffffUL) && (buf[2] <= 0xfd) && (buf[3] >= 2)) {
      if (rid != NULL) *rid = buf[2];
      if (err != NULL) *err = buf[4];
      uint8_t len = buf[3] - 2;
      uint8_t sum = 0;
      if (len <= rdatasize) {
        if (gets (rdata, len) == len) {
          *rlen = len;
          for (int i = 0; i < len; i++) sum += rdata[i];
          if (gets (buf, 1) == 1) return (sum = ~buf[0]);
        }
      }
    }
  }
  *rlen = 0;
  return false;
}

// read byte
bool DXLIB::ReadByteData (uint8_t id, uint16_t addr, uint8_t *data, uint8_t *err) {
  uint8_t buf[2] = {addr, 1};
  if (TxPacket (id, 0x02, buf, 2)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr)) {
      if (err != NULL) *err = rerr;
      if ((len == 1) && (rid == id) && ((rerr & 0x5a) == 0)) {
        *data = buf[0];
        return true;
      }
    }
  }
  return false;
}

// read word
bool DXLIB::ReadWordData (uint8_t id, uint16_t addr, uint16_t *data, uint8_t *err) {
  uint8_t buf[2] = {addr, 2};
  if (TxPacket (id, 0x02, buf, 2)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr)) {
      if (err != NULL) *err = rerr;
      if ((len == 2) && (rid == id) && ((rerr & 0x5a) == 0)) {
        *data = * (uint16_t *)&buf[0];
        return true;
      }
    }
  }
  return false;
}

// read long
bool DXLIB::ReadLongData (uint8_t id, uint16_t addr, uint32_t *data, uint8_t *err) {
  uint8_t buf[4] = {addr, 4};
  if (TxPacket (id, 0x02, buf, 2)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr)) {
      if (err != NULL) *err = rerr;
      if ((len == 4) && (rid == id) && ((rerr & 0x5a) == 0)) {
        *data = * (uint32_t *)&buf[0];
        return true;
      }
    }
  }
  return false;
}

// write byte
bool DXLIB::WriteByteData (uint8_t id, uint16_t addr, uint8_t data, uint8_t *err) {
  uint8_t buf[2] = {addr, data};
  bool result = false;
  if (TxPacket (id, 0x03, buf, 2)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len,  &rerr))
      result = (rid == id) && (len == 0) && ((rerr & 0x5a) == 0);
    if (err != NULL) *err = rerr;
  }
  return result;
}

// write word
bool DXLIB::WriteWordData (uint8_t id, uint16_t addr, uint16_t data, uint8_t *err) {
  uint8_t buf[3] = {addr, data & 0xff, data >> 8};
  bool result = false;
  if (TxPacket (id, 0x03, buf, 3)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr))
      result = (rid == id) && (len == 0) && ((rerr & 0x5a) == 0);
    if (err != NULL) *err = rerr;
  }
  return result;
}

// write long
bool DXLIB::WriteLongData (uint8_t id, uint16_t addr, uint32_t data, uint8_t *err) {
  uint8_t buf[5] = {addr};
  bool result = false;
  * (uint32_t *)&buf[1] = data;
  if (TxPacket (id, 0x03, buf, 5)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr))
      result = (rid == id) && (len == 0) && ((rerr & 0x5a) == 0);
    if (err != NULL) *err = rerr;
  }
  return result;
}

// sync write byte
bool DXLIB::WriteSyncByteData (const uint8_t *pid, uint16_t addr, const uint8_t *pdata, int num) {
  if (!active()) return false;
  uint8_t buf[7] = {0xff, 0xff, 0xfe, (uint8_t) ((num * 2 + 4) & 0xff), 0x83, addr, 1};
  uint8_t sum = 0;
  for (int i = 2; i < 7; i++) sum += buf[i];
  PurgeRXBuff();
  puts (buf, 7);
  for (int i = 0; i < num; i++) {
    sum += pid[i];
    putch (pid[i]);
    sum += pdata[i];
    putch (pdata[i]);
  }
  putch (~sum);
  flush();
  return true;
}

// sync write word
bool DXLIB::WriteSyncWordData (const uint8_t *pid, uint16_t addr, const uint16_t *pdata, int num) {
  if (!active()) return false;
  uint8_t buf[7] = {0xff, 0xff, 0xfe, (uint8_t) ((num * 3 + 4) & 0xff), 0x83, addr, 2};
  uint8_t sum = 0;
  for (int i = 2; i < 7; i++) sum += buf[i];
  PurgeRXBuff();
  puts (buf, 7);
  for (int i = 0; i < num; i++) {
    sum += pid[i];
    putch (pid[i]);
    sum += (pdata[i] & 0xff);
    putch ((pdata[i] & 0xff));
    sum += ((pdata[i] >> 8) & 0xff);
    putch ((pdata[i] >> 8) & 0xff);
  }
  putch (~sum);
  flush();
  return true;
}

// sync write long
bool DXLIB::WriteSyncLongData (const uint8_t *pid, uint16_t addr, const uint32_t *pdata, int num) {
  if (!active()) return false;
  uint8_t buf[7] = {0xff, 0xff, 0xfe, (uint8_t) ((num * 5 + 4) & 0xff), 0x83, addr, 4};
  uint8_t sum = 0;
  for (int i = 2; i < 7; i++) sum += buf[i];
  PurgeRXBuff();
  puts (buf, 7);
  for (int i = 0; i < num; i++) {
    sum += pid[i];
    putch (pid[i]);
    sum += (pdata[i] & 0xff);
    putch ((pdata[i] & 0xff));
    sum += ((pdata[i] >> 8) & 0xff);
    putch ((pdata[i] >> 8) & 0xff);
    sum += ((pdata[i] >> 16) & 0xff);
    putch ((pdata[i] >> 16) & 0xff);
    sum += ((pdata[i] >> 24) & 0xff);
    putch ((pdata[i] >> 24) & 0xff);
  }
  putch (~sum);
  flush();
  return true;
}

// ping
bool DXLIB::Ping (uint8_t id, uint8_t *err) {
  if (id > 253) return false;
  uint8_t buf[5];
  bool result = false;
  if (TxPacket (id, 0x01, NULL, 0)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr))
      result = (rid == id) && (len == 0) && ((rerr & 0x5a) == 0);
    if (err != NULL) *err = rerr;
  }
  return result;
}

// factory reset
bool DXLIB::Reset (uint8_t id, uint8_t *err) {
  if (id >= 254) return false;
  uint8_t buf[5];
  bool result = false;
  if (TxPacket (id, 0x06, NULL, 0)) {
    uint8_t rid, rerr;
    uint16_t len;
    if (RxPacket (buf, sizeof (buf), &rid, &len, &rerr))
      result = (rid == id) && (len == 0) && ((rerr & 0x5a) == 0);
    if (err != NULL) *err = rerr;
  }
  return result;
}
